from dataclasses import dataclass
from pathlib import Path
from zipfile import ZipFile
from typing import List, Literal, Optional, TypeVar, Generic, Union, Callable
import orjson

from sm.misc.funcs import batch, get_latest_path
from sm.misc.matrix import Matrix
from sm.outputs.semantic_model import SemanticModel
from sm.inputs.prelude import ColumnBasedTable, Context, Link
from slugify import slugify
from serde import json
import shutil
from urllib.parse import urlparse
from hashlib import md5

T = TypeVar("T")


@dataclass
class Example(Generic[T]):
    sms: List[SemanticModel]
    table: T


@dataclass
class FullTable:
    table: ColumnBasedTable
    context: Context
    links: Matrix[List[Link]]

    def keep_rows(self, row_index: List[int]):
        """Keep only the rows in the table that are in row_index."""
        self.links.data = [self.links.data[i] for i in row_index]
        self.table._df = None
        for col in self.table.columns:
            col.values = [col.values[i] for i in row_index]

    def to_dict(self):
        return {
            "version": 2,
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in cell] for cell in row]
                for row in self.links.data
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        version = obj["version"]
        if not (version == "1.2" or version == "1.1" or version == 2):
            raise ValueError(f"Unknown version: {version}")

        return cls(
            table=ColumnBasedTable.from_dict(obj["table"]),
            context=Context.from_dict(obj["context"]),
            links=Matrix(
                [
                    [[Link.from_dict(link) for link in cell] for cell in row]
                    for row in obj["links"]
                ]
            ),
        )


@dataclass
class Dataset:
    location: Path

    @property
    def description_dir(self) -> Path:
        return self.location / "descriptions"

    @property
    def table_dir(self) -> Path:
        return self.location / "tables"

    def load(self) -> List[Example[FullTable]]:
        """Load dataset from a folder. Assuming the following structure:

        descriptions (containing semantic descriptions of tables)
        ├── <table_fs_id>
        │   ├── version.01.json
        │   ├── version.02.json
        │   └── ...
            or
        ├── <table_fs_id>.json
        ├── ...
        tables (containing list of tables, the type of table depends on )
        ├── <table_fs_id>.json[.gz|.bz2|.lz4]
        ├── ...

        We also support compressing formats such as .zip.
        descriptions
        ├── part-<num>.zip (no nested version folders)
        │   ├── <table_fs_id>.json
        |   |   ...
        tables
        ├── part-<num>.zip
        │   ├── <table_fs_id>.json
        |   |   ...

        Args:
            data_dir:

        Returns:

        """
        descdir = self.description_dir
        tabledir = self.table_dir

        examples = []
        for infile in sorted(tabledir.iterdir()):
            suffixes = infile.suffixes
            if infile.name.startswith(".") or len(suffixes) == 0:
                continue

            if suffixes[0] == ".json":
                example_id = infile.name[: -sum(len(x) for x in suffixes)]
                table = json.deser(infile, FullTable)

                if (descdir / example_id).exists():
                    desc_file = get_latest_path(descdir / f"{example_id}/version.json")
                    assert desc_file is not None
                else:
                    desc_file = descdir / f"{example_id}.json"
                    assert desc_file.exists()

                raw_sms = json.deser(desc_file)
                sms = [SemanticModel.from_dict(sm) for sm in raw_sms]

                examples.append(Example(sms=sms, table=table))
            elif infile.name.endswith(".zip"):

                part = {}
                with ZipFile(infile, mode="r") as zf:
                    for file in zf.infolist():
                        if not file.filename.endswith(".json"):
                            continue

                        table_id = Path(file.filename).stem
                        with zf.open(file, mode="r") as f:
                            table = FullTable.from_dict(orjson.loads(f.read()))
                        part[table_id] = table

                lst = []
                with ZipFile(descdir / infile.name, mode="r") as zf:
                    for file in zf.infolist():
                        table_id = Path(file.filename).stem
                        if table_id not in part:
                            continue

                        assert file.filename.endswith(".json")
                        with zf.open(file, mode="r") as f:
                            sms = [
                                SemanticModel.from_dict(sm)
                                for sm in orjson.loads(f.read())
                            ]
                            lst.append(Example(sms=sms, table=part[table_id]))

                assert len(lst) == len(part)
                examples.extend(lst)

        return examples

    def save(
        self,
        examples: List[Example[FullTable]],
        individual_table_compressed: Optional[Literal["gz", "bz2", "lz4"]] = None,
        batch_compressed: bool = False,
        batch_size: int = 100,
        table_fmt_indent: Literal[0, 2] = 0,
        clean_previous_data: bool = True,
    ):
        descdir = self.description_dir
        if descdir.exists() and clean_previous_data:
            shutil.rmtree(descdir)
        descdir.mkdir(parents=True, exist_ok=True)

        tabledir = self.table_dir
        if tabledir.exists() and clean_previous_data:
            shutil.rmtree(tabledir)
        tabledir.mkdir(parents=True, exist_ok=True)

        if batch_compressed:
            for i, bexamples in enumerate(batch(batch_size, examples)):
                bexamples: List[Example[FullTable]]
                filename = f"part-{i:04d}.zip"
                with ZipFile(descdir / filename, "w") as dzf, ZipFile(
                    tabledir / filename, "w"
                ) as tzf:
                    for e in bexamples:
                        ename = get_friendly_fs_id(e.table.table.table_id) + ".json"
                        dzf.writestr(
                            ename,
                            data=orjson.dumps([sm.to_dict() for sm in e.sms]),
                        )
                        tzf.writestr(
                            ename,
                            data=orjson.dumps(e.table.to_dict()),
                        )
        else:
            for e in examples:
                filename = get_friendly_fs_id(e.table.table.table_id)
                (descdir / filename).mkdir(exist_ok=True)
                compressed_filename = (
                    filename + f".json.{individual_table_compressed}"
                    if individual_table_compressed is not None
                    else filename + ".json"
                )
                json.ser(
                    [sm.to_dict() for sm in e.sms],
                    descdir / filename / "version.01.json",
                    indent=2,
                )
                json.ser(
                    e.table, tabledir / compressed_filename, indent=table_fmt_indent
                )


def get_friendly_fs_id(id: str) -> str:
    if id.startswith("http://") or id.startswith("https://"):
        if id.find("dbpedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        if id.find("wikipedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        raise NotImplementedError()
    return slugify(id.replace("/", "_"), lowercase=False).replace("-", "_")



from __future__ import annotations
from typing import List, Optional


class EntityId(str):
    __slots__ = ("type",)
    type: str

    def __new__(cls, id: str, type: str):
        obj = str.__new__(cls, id)
        obj.type = type
        return obj

    def to_dict(self) -> dict[str, str]:
        return {
            "id": self,
            "type": self.type,
        }

    @staticmethod
    def from_dict(obj: dict) -> EntityId:
        return EntityId(
            id=obj["id"],
            type=obj["type"],
        )

    def __getnewargs__(self) -> tuple[str, str]:
        return str(self), self.type


WIKIDATA = "wikidata"
WIKIDATA_NIL_ENTITY = EntityId("Q0", WIKIDATA)


class Link:
    __slots__ = ("start", "end", "url", "entities")
    """Represent a link in a cell, a link may not cover the whole cell, so a cell
    may have multiple links.

    Attributes:
        start: start index of the link in the cell
        end: end index of the link in the cell
        url: url of the link, none means there is no hyperlink
        entities: entities linked by the link, each entity is from each knowledge graph.
            If entities is empty, it means the link should not link to any entity.
            If an entity of a target KG is NIL, it means the link should link to NIL entity
            because there is no corresponding entity in that knowledge graph.
    """

    def __init__(
        self, start: int, end: int, url: Optional[str], entities: List[EntityId]
    ):
        self.start = start
        self.end = end  # exclusive
        self.url = url  # url of the link, none means there is no hyperlink
        self.entities = entities

    def to_dict(self):
        return {
            "version": 2,
            "start": self.start,
            "end": self.end,
            "url": self.url,
            "entities": [e.to_dict() for e in self.entities],
        }

    @staticmethod
    def from_dict(obj: dict):
        version = obj.get("version")
        if version == 2:
            return Link(
                start=obj["start"],
                end=obj["end"],
                url=obj["url"],
                entities=[EntityId.from_dict(e) for e in obj["entities"]],
            )
        if version is None:
            return Link(
                start=obj["start"],
                end=obj["end"],
                url=obj["url"],
                entities=[EntityId(id=eid, type=WIKIDATA)]
                if (eid := obj["entity_id"]) is not None
                else [],
            )
        raise ValueError(f"Unknown version: {version}")



import ray
from typing import (
    TypeVar,
    Union,
    Callable,
    List,
    Optional,
)

from loguru import logger
from tqdm.auto import tqdm

R = TypeVar("R")
OBJECTS = {}


def ray_init(**kwargs):
    if not ray.is_initialized():
        ray.init(**kwargs)


def ray_map(
    remote_fn: Callable[..., "ray.ObjectRef[R]"],
    args_lst: List[Union[list, tuple]],
    verbose: bool = False,
    poll_interval: float = 0.1,
    concurrent_submissions: int = 3000,
    desc: Optional[str] = None,
) -> List[R]:
    n_jobs = len(args_lst)

    with tqdm(total=n_jobs, desc=desc, disable=not verbose) as pbar:
        output: List[R] = [None] * n_jobs  # type: ignore

        notready_refs = []
        ref2index = {}
        for i, args in enumerate(args_lst):
            # submit a task and add it to not ready queue and ref2index
            ref = remote_fn(*args)
            notready_refs.append(ref)
            ref2index[ref] = i

            # when the not ready queue is full, wait for some tasks to finish
            while len(notready_refs) >= concurrent_submissions:
                ready_refs, notready_refs = ray.wait(
                    notready_refs, timeout=poll_interval
                )
                pbar.update(len(ready_refs))
                for ref in ready_refs:
                    output[ref2index[ref]] = ray.get(ref)

        while len(notready_refs) > 0:
            ready_refs, notready_refs = ray.wait(notready_refs, timeout=poll_interval)
            pbar.update(len(ready_refs))
            for ref in ready_refs:
                output[ref2index[ref]] = ray.get(ref)

        return output


def get_instance(constructor: Callable[[], R], name: Optional[str] = None) -> R:
    """A utility function to get a singleton, which can be created from the given constructor.

    One use case of this function is we have a big object that is expensive to send
    to individual task repeatedly. If the process are retrieved from a pool,
    this allows us to create the object per process instead of per task.
    """
    global OBJECTS

    if name is None:
        assert (
            constructor.__name__ != "<lambda>"
        ), "Cannot use lambda as a name because it will keep changing"
        name = constructor  # type: ignore

    if name not in OBJECTS:
        logger.trace("Create a new instance of {}", name)
        OBJECTS[name] = constructor()
    return OBJECTS[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sm', 'sm.evaluation', 'sm.inputs', 'sm.misc', 'sm.namespaces', 'sm.outputs']

package_data = \
{'': ['*'], 'sm': ['data/*']}

install_requires = \
['Deprecated>=1.2.13,<2.0.0',
 'colorama>=0.4.4',
 'graph-wrapper>=1.5.0,<2.0.0',
 'ipython>=8.0.1,<9.0.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.3,<4.0.0',
 'orjson>=3.8.2,<4.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'pyrsistent>=0.17.3,<0.18.0',
 'python-slugify>=6.1.2,<7.0.0',
 'ray>=2.0.1,<3.0.0',
 'rdflib>=6.1.1,<7.0.0',
 'rsoup>=2.5.1,<3.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'serde2[all]>=1.5.0,<2.0.0',
 'timer4>=1.0.2,<2.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'ujson>=5.5.0,<6.0.0']

setup_kwargs = {
    'name': 'sem-desc',
    'version': '4.2.0',
    'description': 'Package providing basic functionalities for the semantic modeling problem',
    'long_description': '# SEM-DESC ![PyPI](https://img.shields.io/pypi/v/sem-desc)\n\nContaining basic functions (input, output, dataset, evaluation metrics) for the semantic modeling problem.\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/binh-vu/sm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
